#
# Copyright 2021, by the Max Planck Institute for Gravitational Physics
# (Albert Einstein Institute). ALL RIGHTS RESERVED.
#
"""
Trimmer dor LISANode data.

This script takes HDF5 data produced by LISANode and removes unnecessary datasets.
"""

import sys

import h5py

KEEP_KEYS = [
    "ranging1",
    "ranging1_p",
    "ranging2",
    "ranging2_p",
    "ranging3",
    "ranging3_p",
    "s_c_fluctuations1",
    "s_c_fluctuations1_p",
    "s_c_fluctuations2",
    "s_c_fluctuations2_p",
    "s_c_fluctuations3",
    "s_c_fluctuations3_p",
    "tm_c_fluctuations1",
    "tm_c_fluctuations1_p",
    "tm_c_fluctuations2",
    "tm_c_fluctuations2_p",
    "tm_c_fluctuations3",
    "tm_c_fluctuations3_p",
    "ref_c_fluctuations1",
    "ref_c_fluctuations1_p",
    "ref_c_fluctuations2",
    "ref_c_fluctuations2_p",
    "ref_c_fluctuations3",
    "ref_c_fluctuations3_p",
]
LENGTH = 400000


if __name__ == "__main__":
    path_in = str(sys.argv[1])
    path_out = str(sys.argv[2])
    with h5py.File(path_in, "r") as hdf5_in, h5py.File(path_out, "w") as hdf5_out:
        for key in hdf5_in:
            if key in KEEP_KEYS:
                hdf5_out.create_dataset(key, data=hdf5_in[key][-LENGTH:])



#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Defines TDI first-and-second-generation Sagnac variables.

Authors:
    Martin Staab <martin.staab@aei.mpg.de>
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
"""

from . import core, intervar

#: First-generation :math:`\alpha_1` Sagnac combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
ALPHA1_ETA = core.LISATDICombination.from_string("1321 -1321")

#: First-generation :math:`\beta_1` Sagnac combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
BETA1_ETA = ALPHA1_ETA.rotated()

#: First-generation :math:`\gamma_1` Sagnac combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
GAMMA1_ETA = BETA1_ETA.rotated()

#: First-generation :math:`\zeta_1` fully-symmetric Sagnac combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
ZETA1_ETA = core.LISATDICombination(
    {
        "eta_12": [(1, ["D_23"])],
        "eta_23": [(1, ["D_31"])],
        "eta_31": [(1, ["D_12"])],
        "eta_13": [(-1, ["D_32"])],
        "eta_32": [(-1, ["D_21"])],
        "eta_21": [(-1, ["D_13"])],
    }
)

#: First-generation :math:`\alpha_1` Sagnac combination.
#:
#: This combination is the composition of ``ALPHA1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
ALPHA1 = ALPHA1_ETA @ intervar.ETA_SET

#: First-generation :math:`\beta_1` Sagnac combination.
#:
#: This combination is the composition of ``BETA1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
BETA1 = BETA1_ETA @ intervar.ETA_SET

#: First-generation :math:`\gamma_1` Sagnac combination.
#:
#: This combination is the composition of ``GAMMA1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
GAMMA1 = GAMMA1_ETA @ intervar.ETA_SET

#: First-generation :math:`\zeta_1` fully-symmetric Sagnac combination.
#:
#: This combination is the composition of ``ZETA1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
ZETA1 = ZETA1_ETA @ intervar.ETA_SET

#: Second-generation :math:`\alpha_2` Sagnac combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
ALPHA2_ETA = core.LISATDICombination.from_string("1231321 -1321231")

#: Second-generation :math:`\beta_2` Sagnac combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
BETA2_ETA = ALPHA2_ETA.rotated()

#: Second-generation :math:`\gamma_2` Sagnac combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
GAMMA2_ETA = BETA2_ETA.rotated()

#: Second-generation :math:`\zeta_{21}` fully-symmetric Sagnac comb. of first kind.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
ZETA21_ETA = core.LISATDICombination(
    {
        "eta_12": [(1, ["D_23", "D_32"]), (-1, ["D_13", "D_21", "D_32"])],
        "eta_23": [(1, ["D_32", "D_13"]), (-1, ["D_12", "D_31", "D_13"])],
        "eta_31": [(1, ["D_23", "D_12"]), (-1, ["D_13", "D_21", "D_12"])],
        "eta_13": [(1, ["D_12", "D_31", "D_23"]), (-1, ["D_32", "D_23"])],
        "eta_32": [(1, ["D_13", "D_21", "D_12"]), (-1, ["D_23", "D_12"])],
        "eta_21": [(1, ["D_12", "D_31", "D_13"]), (-1, ["D_32", "D_13"])],
    }
)

#: Second-generation :math:`\zeta_{22}` fully-symmetric Sagnac comb. of second kind.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
ZETA22_ETA = ZETA21_ETA.rotated()

#: Second-generation :math:`\zeta_{23}` fully-symmetric Sagnac comb. of third kind.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
ZETA23_ETA = ZETA22_ETA.rotated()

#: Second-generation :math:`\alpha_2` Sagnac combination.
#:
#: This combination is the composition of ``ALPHA1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
ALPHA2 = ALPHA2_ETA @ intervar.ETA_SET

#: Second-generation :math:`\beta_2` Sagnac combination.
#:
#: This combination is the composition of ``BETA1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
BETA2 = BETA2_ETA @ intervar.ETA_SET

#: Second-generation :math:`\gamma_2` Sagnac combination.
#:
#: This combination is the composition of ``GAMMA1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
GAMMA2 = GAMMA2_ETA @ intervar.ETA_SET

#: Second-generation :math:`\zeta_{21}` fully-symmetric Sagnac comb. of first kind.
#:
#: This combination is the composition of ``ZETA21_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
ZETA21 = ZETA21_ETA @ intervar.ETA_SET

#: Second-generation :math:`\zeta_{22}` fully-symmetric Sagnac comb. of second kind.
#:
#: This combination is the composition of ``ZETA21_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
ZETA22 = ZETA22_ETA @ intervar.ETA_SET

#: Second-generation :math:`\zeta_{23}` fully-symmetric Sagnac comb. of third kind.
#:
#: This combination is the composition of ``ZETA21_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
ZETA23 = ZETA23_ETA @ intervar.ETA_SET


def compute_factorized_sagnac(
    data, etas=None, *, rot=0, order=45, generation=2, unit="frequency"
):
    r"""Numerically optimized computation of the Sagnac variables.

    These combinations can be factorized and evaluated in multiple steps. This
    reduces computation time (less delays to compute) and in some instances
    also improves numerical precision if intermediary results are already small.

    First-generation Sagnac combinations only use each :math:`\eta` once,
    so they cannot be factorized. However, the :math:`\eta_{ij}` can be
    pre-computed,

    .. math::

        \alpha_1 ={}&
        (\eta_{12} + \delay_{12} \eta_{23} + \delay_{123} \eta_{31}) \\
        - (\eta_{13} + \delay_{13} \eta_{32} + \delay_{132} \eta_{21}).

    For second-generation Sagnac combinations, we can instead write

    .. math::

        alpha2 ={}& (1 - \delay_{1321})
        (\eta_{12} + \delay_{12} \eta_{23} + \delay_{123} \eta_{31}) \\
        &- (1 - \delay_{1231})
        (\eta_{13} + \delay_{13} \eta_{32} + \delay_{132} \eta_{21}).

    We first define the single circle round-trips,

    .. math::

        \text{beam}_1 &= \eta_{12} + \delay_{12} \eta_{23} + \delay_{123} \eta_{31}, \\
        \text{beam}_2 &= \eta_{13} + \delay_{13} \eta_{32} + \delay_{132} \eta_{21}.

    In the case of :math:`\alpha_1`, we directly compute the result

    .. math:: \text{beam}_1 - \text{beam}_2,

    and for :math:`\alpha_2`,

    .. math:: (1 - \delay_{1321}) \text{beam}_1 - (1 - \delay_{1231}) \text{beam}_2.

    Note that the signs in these combinations are chosen to agree with the
    existing conventions in PyTDI and the LDC.

    Args:
        data (pytdi.interface.Data): data object
        etas (np.ndarray): pre-computed :math:`\eta_{ij}` intermediary variables
        rot (int): rotation of indices for :math:`\alpha, \beta, \gamma`
        order (int): interpolation order
        unit (str): unit of input data, either 'frequency' or 'phase'

    Returns:
        (np.ndarray) Evaluated TDI Sagnac combination.
    """
    # Define rotations of the variables
    # We can't rely on PyTDIs built-in rotation as
    # the variable names are incompatible
    if rot == 0:
        links = ["12", "21", "23", "32", "31", "13"]
    elif rot == 1:
        links = ["23", "32", "31", "13", "12", "21"]
    elif rot == 2:
        links = ["31", "13", "12", "21", "23", "32"]
    else:
        raise ValueError(f"rot must be 0, 1 or 2 (got {rot})")

    # IFOs measuring single transponder
    # Use pre-computed etas if provided
    if etas is not None:
        beam_1 = core.LISATDICombination(
            {
                f"eta_{links[0]}": [(1, [])],
                f"eta_{links[2]}": [(1, [f"D_{links[0]}"])],
                f"eta_{links[4]}": [(1, [f"D_{links[0]}", f"D_{links[2]}"])],
            }
        ).build(**data.args)(etas, order=order, unit=unit)
        beam_2 = core.LISATDICombination(
            {
                f"eta_{links[5]}": [(1, [])],
                f"eta_{links[3]}": [(1, [f"D_{links[5]}"])],
                f"eta_{links[1]}": [(1, [f"D_{links[5]}", f"D_{links[3]}"])],
            }
        ).build(**data.args)(etas, order=order, unit=unit)
    else:
        # Use ETA_SET to compute result from IFO measurements
        beam_1 = (
            core.LISATDICombination(
                {
                    f"eta_{links[0]}": [(1, [])],
                    f"eta_{links[2]}": [(1, [f"D_{links[0]}"])],
                    f"eta_{links[4]}": [(1, [f"D_{links[0]}", f"D_{links[2]}"])],
                }
            )
            @ intervar.ETA_SET
        ).build(**data.args)(data.measurements, order=order, unit=unit)
        beam_2 = (
            core.LISATDICombination(
                {
                    f"eta_{links[5]}": [(1, [])],
                    f"eta_{links[3]}": [(1, [f"D_{links[5]}"])],
                    f"eta_{links[1]}": [(1, [f"D_{links[5]}", f"D_{links[3]}"])],
                }
            )
            @ intervar.ETA_SET
        ).build(**data.args)(data.measurements, order=order, unit=unit)

    if generation == 1:
        # Full combination
        result = beam_1 - beam_2
    elif generation == 2:
        # Full combination
        result = core.LISATDICombination(
            {
                "beam_1": [
                    (1, []),
                    (-1, [f"D_{links[5]}", f"D_{links[3]}", f"D_{links[1]}"]),
                ],
                "beam_2": [
                    (-1, []),
                    (1, [f"D_{links[0]}", f"D_{links[2]}", f"D_{links[4]}"]),
                ],
            }
        ).build(**data.args)(
            {"beam_1": beam_1, "beam_2": beam_2}, order=order, unit=unit
        )
    else:
        raise ValueError(f"invalid generation '{generation}', must be 1 or 2")

    return result



#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Configure test environment.
"""


def pytest_addoption(parser):
    """Adds custome command line options."""
    parser.addoption("--plotting", action="store_true", default=False)


def pytest_generate_tests(metafunc):
    """Parameterizes tests."""
    if "plotting" in metafunc.fixturenames:
        if metafunc.config.getoption("plotting"):
            metafunc.parametrize("plotting", (True,))
        else:
            metafunc.parametrize("plotting", (False,))



"""PyTDI module."""

from importlib.metadata import PackageNotFoundError, metadata, version

from .core import LISAClockCorrection, LISATDICombination, TDICombination
from .interface import Data

try:
    __version__ = version(__name__)
    __license__ = metadata(__name__)["license"]
except PackageNotFoundError:
    __version__ = "0.0.0"  # Fallback for development mode
    __license__ = "Undefined"

__copyright__ = (
    "2021, Max Planck Institute for Gravitational Physics "
    "(Albert Einstein Institute) and California Institute of Technology"
)



#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Test that orthogonal combinations (AET) can reduce laser noise (using LISANode data).
"""

from integration import BASE_PATH, psd_threshold, read_from_lisanode

import pytdi.intervar
import pytdi.ortho


def test_ortho_1st_gen():
    """Test that the first-generation AET combinations reduce laser noise."""
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-keplerianorbits.h5"
    )
    combinations = {"A1": pytdi.ortho.A1, "E1": pytdi.ortho.E1, "T1": pytdi.ortho.T1}
    for name, tdi in combinations.items():
        data_x = tdi.build(delays, 3.2)(data_ln, unit="phase")
        psd_threshold(data_x[200:], 1e-39, freq_range=(1e-4, 8e-1), fs=3.2, name=name)


def test_factorized_ortho_1st_gen():
    """Test that the factorized first-generation AET combinations reduce laser noise."""
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-keplerianorbits.h5"
    )
    data = pytdi.interface.Data(data_ln, delays, fs=3.2)
    ortho = pytdi.ortho.compute_factorized_ortho(data, generation=1, unit="phase")
    psd_threshold(
        ortho["A"][200:], 1e-39, freq_range=(1e-4, 8e-1), fs=3.2, name="facto-A1"
    )
    psd_threshold(
        ortho["E"][200:], 1e-39, freq_range=(1e-4, 8e-1), fs=3.2, name="facto-E1"
    )
    psd_threshold(
        ortho["T"][200:], 1e-39, freq_range=(1e-4, 8e-1), fs=3.2, name="facto-T1"
    )


def test_ortho_2nd_gen():
    """Test that the second-generation AET combinations reduce laser noise."""
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-keplerianorbits.h5"
    )
    combinations = {"A2": pytdi.ortho.A2, "E2": pytdi.ortho.E2, "T2": pytdi.ortho.T2}
    for name, tdi in combinations.items():
        data_x = tdi.build(delays, 3.2)(data_ln, unit="phase")
        psd_threshold(data_x[500:], 1e-41, freq_range=(1e-4, 6e-1), fs=3.2, name=name)


def test_factorized_ortho_2nd_gen():
    """Test that the factorized second-generation AET combinations reduce laser noise."""
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-keplerianorbits.h5"
    )
    data = pytdi.interface.Data(data_ln, delays, fs=3.2)
    ortho = pytdi.ortho.compute_factorized_ortho(data, generation=2, unit="phase")
    psd_threshold(
        ortho["A"][500:], 1e-41, freq_range=(1e-4, 6e-1), fs=3.2, name="facto-A2"
    )
    psd_threshold(
        ortho["E"][500:], 1e-41, freq_range=(1e-4, 6e-1), fs=3.2, name="facto-E2"
    )
    psd_threshold(
        ortho["T"][500:], 1e-41, freq_range=(1e-4, 6e-1), fs=3.2, name="facto-T2"
    )



#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Defines TDI intermediary variables ξ (xi) and η (eta).

Authors:
    Martin Staab <martin.staab@aei.mpg.de>
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
    Olaf Hartwig <olaf.hartwig@aei.mpg.de>
"""

from . import core

#: :math:`\xi_{12}` spacecraft jitter-reducing intermediary variable.
XI_12 = core.LISATDICombination(
    {
        "isi_12": [(1.0, [])],
        "rfi_12": [(0.5, [])],
        "rfi_21": [(0.5, ["D_12"])],
        "tmi_12": [(-0.5, [])],
        "tmi_21": [(-0.5, ["D_12"])],
    }
)

#: :math:`\xi_{23}` spacecraft jitter-reducing intermediary variable.
XI_23 = XI_12.rotated()

#: :math:`\xi_{31}` spacecraft jitter-reducing intermediary variable.
XI_31 = XI_23.rotated()

#: :math:`\xi_{13}` spacecraft jitter-reducing intermediary variable.
XI_13 = XI_12.reflected(1)

#: :math:`\xi_{21}` spacecraft jitter-reducing intermediary variable.
XI_21 = XI_23.reflected(2)

#: :math:`\xi_{31}` spacecraft jitter-reducing intermediary variable.
XI_32 = XI_31.reflected(3)

#: dict: Dictionary of all 6 :math:`\xi_{ij}` variables.
#:
#: This dictionary is intended to be used for composition of combinations.
XI_SET = {
    "xi_12": XI_12,
    "xi_23": XI_23,
    "xi_31": XI_31,
    "xi_13": XI_13,
    "xi_21": XI_21,
    "xi_32": XI_32,
}

#: :math:`\eta_{12}` intermediary variable to reduce the problem to 3 lasers.
#:
#: This combination is defined as a function of :math:`\xi` measurements in
#: addition to the reference beatnotes.
ETA_12_XI = core.LISATDICombination(
    {
        "xi_12": [(1.0, [])],
        "rfi_23": [(-0.5, ["D_12"])],
        "rfi_21": [(0.5, ["D_12"])],
    },
    allow_reflections=False,
)

#: :math:`\eta_{23}` intermediary variable to reduce the problem to 3 lasers.
#:
#: This combination is defined as a function of :math:`\xi` measurements in
#: addition to the reference beatnotes.
ETA_23_XI = ETA_12_XI.rotated()

#: :math:`\eta_{31}` intermediary variable to reduce the problem to 3 lasers.
#:
#: This combination is defined as a function of :math:`\xi` measurements in
#: addition to the reference beatnotes.
ETA_31_XI = ETA_23_XI.rotated()

#: :math:`\eta_{13}` intermediary variable to reduce the problem to 3 lasers.
#:
#: This combination is defined as a function of :math:`\xi` measurements in
#: addition to the reference beatnotes.
ETA_13_XI = core.LISATDICombination(
    {
        "xi_13": [(1.0, [])],
        "rfi_12": [(0.5, [])],
        "rfi_13": [(-0.5, [])],
    },
    allow_reflections=False,
)

#: :math:`\eta_{21}` intermediary variable to reduce the problem to 3 lasers.
#:
#: This combination is defined as a function of :math:`\xi` measurements in
#: addition to the reference beatnotes.
ETA_21_XI = ETA_13_XI.rotated()

#: :math:`\eta_{32}` intermediary variable to reduce the problem to 3 lasers.
#:
#: This combination is defined as a function of :math:`\xi` measurements in
#: addition to the reference beatnotes.
ETA_32_XI = ETA_21_XI.rotated()

#: dict: Dictionary of all 6 :math:`\eta_{ij}` variables.
#:
#: The :math:`\eta_{ij}` are defined as a function of the :math:`\xi` measurements in
#: addition to the reference beatnotes.
#:
#: This dictionary is intended to be used for composition of combinations.
ETA_XI_SET = {
    "eta_12": ETA_12_XI,
    "eta_23": ETA_23_XI,
    "eta_31": ETA_31_XI,
    "eta_13": ETA_13_XI,
    "eta_21": ETA_21_XI,
    "eta_32": ETA_32_XI,
}

#: dict: Dictionary of all 6 :math:`\eta_{ij}` variables.
#:
#: The :math:`\eta_{ij}` are defined as a function of the beatnotes.
#:
#: This dictionary is intended to be used for composition of combinations.
ETA_SET = {
    "eta_12": ETA_12_XI @ XI_SET,
    "eta_23": ETA_23_XI @ XI_SET,
    "eta_31": ETA_31_XI @ XI_SET,
    "eta_13": ETA_13_XI @ XI_SET,
    "eta_21": ETA_21_XI @ XI_SET,
    "eta_32": ETA_32_XI @ XI_SET,
}


def compute_etas(data, order=45, unit="frequency"):
    r"""Evaluate intermediary variables :math:`\eta_{ij}`.

    This can be used as an alternative to concatenation, to evaluate numerically
    intermediary variables and use them as input to construct other variables.
    It might save time if the same intermediary variables are used multiple times.

    Args:
        data (pytdi.interface.Data): data object
        order (int): order of interpolation
        unit (str): unit of input data, either 'frequency' or 'phase'
    """
    return {
        key: eta.build(**data.args)(data.measurements, order=order, unit=unit)
        for key, eta in ETA_SET.items()
    }



#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Test PyTDI's interfaces to other modules through the :class:`pytdi.Data` class.
"""

from h5py import File
from pytest import approx

from pytdi import Data
from pytdi.michelson import X1


def test_gw_file_1_1():
    """Test that we can read GW files v1.1 with orbit files v1.0.2.

    Test GW file can be generated using LISA GW Response and the following script.

        from h5py import File
        from lisagwresponse import GalacticBinary

        orbits = 'tests/keplerian-orbits-1-0-2.h5'
        with File(orbits, 'r') as f:
            t0 = f.attrs['t0'] + 10

        galbin = GalacticBinary(
            A=1.0,
            f=1E-3,
            orbits=orbits,
            t0=t0,
            size=100,
            gw_beta=0, gw_lambda=0,
        )

        galbin.write('tests/gws-1-1.h5')
    """
    gw_file = "tests/gws-1-1.h5"
    orbit_file = "tests/keplerian-orbits-1-0-2.h5"

    with File(gw_file, "r") as gwf:

        data = Data.from_gws(gw_file, orbit_file)
        X1.build(**data.args)(data.measurements)
        assert data.fs == gwf.attrs["fs"]
        for mosa in Data.MOSAS:
            assert data.measurements[f"isi_{mosa}"] != approx(0.0)
            assert data.measurements[f"isi_sb_{mosa}"] != approx(0.0)
            assert data.measurements[f"tmi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_sb_{mosa}"] == approx(0.0)

        data = Data.from_gws(gw_file, orbit_file, gw_dataset="tcb")
        X1.build(**data.args)(data.measurements)
        assert data.fs == gwf.attrs["fs"]
        for mosa in Data.MOSAS:
            assert data.measurements[f"isi_{mosa}"] != approx(0.0)
            assert data.measurements[f"isi_sb_{mosa}"] != approx(0.0)
            assert data.measurements[f"tmi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_sb_{mosa}"] == approx(0.0)

        data = Data.from_gws(gw_file, orbit_file, gw_dataset="tps")
        X1.build(**data.args)(data.measurements)
        assert data.fs == gwf.attrs["fs"]
        for mosa in Data.MOSAS:
            assert data.measurements[f"isi_{mosa}"] != approx(0.0)
            assert data.measurements[f"isi_sb_{mosa}"] != approx(0.0)
            assert data.measurements[f"tmi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_sb_{mosa}"] == approx(0.0)


def test_gw_file_2_0():
    """Test that we can read GW files v2.0 with orbit files v2.0.

    Test GW file can be generated using LISA GW Response and the following script.

        from h5py import File
        from lisagwresponse import GalacticBinary

        orbits = 'tests/keplerian-orbits-2-0.h5'
        with File(orbits, 'r') as f:
            t0 = f.attrs['t0'] + 10

        galbin = GalacticBinary(
            A=1.0,
            f=1E-3,
            orbits=orbits,
            t0=t0,
            size=100,
            gw_beta=0, gw_lambda=0,
        )

        galbin.write('tests/gws-2-0.h5')
    """
    gw_file = "tests/gws-2-0.h5"
    orbit_file = "tests/keplerian-orbits-2-0.h5"

    with File(gw_file, "r") as gwf:

        data = Data.from_gws(gw_file, orbit_file)
        X1.build(**data.args)(data.measurements)
        assert data.fs == gwf.attrs["fs"]
        for mosa in Data.MOSAS:
            assert data.measurements[f"isi_{mosa}"] != approx(0.0)
            assert data.measurements[f"isi_sb_{mosa}"] != approx(0.0)
            assert data.measurements[f"tmi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_sb_{mosa}"] == approx(0.0)

        data = Data.from_gws(gw_file, orbit_file, gw_dataset="tcb")
        X1.build(**data.args)(data.measurements)
        assert data.fs == gwf.attrs["fs"]
        for mosa in Data.MOSAS:
            assert data.measurements[f"isi_{mosa}"] != approx(0.0)
            assert data.measurements[f"isi_sb_{mosa}"] != approx(0.0)
            assert data.measurements[f"tmi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_sb_{mosa}"] == approx(0.0)

        data = Data.from_gws(gw_file, orbit_file, gw_dataset="tps")
        X1.build(**data.args)(data.measurements)
        assert data.fs == gwf.attrs["fs"]
        for mosa in Data.MOSAS:
            assert data.measurements[f"isi_{mosa}"] != approx(0.0)
            assert data.measurements[f"isi_sb_{mosa}"] != approx(0.0)
            assert data.measurements[f"tmi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_sb_{mosa}"] == approx(0.0)


def test_gw_file_2_3():
    """Test that we can read GW files v2.3 with orbit files v2.3.

    Test GW file can be generated using LISA GW Response and the following script.

        from h5py import File
        from lisagwresponse import GalacticBinary

        orbits = 'tests/keplerian-orbits-2-3.h5'
        with File(orbits, 'r') as f:
            t0 = f.attrs['t0'] + 10

        galbin = GalacticBinary(
            A=1.0,
            f=1E-3,
            orbits=orbits,
            t0=t0,
            size=100,
            gw_beta=0, gw_lambda=0,
        )

        galbin.write('tests/gws-2-3.h5')
    """
    gw_file = "tests/gws-2-3.h5"
    orbit_file = "tests/keplerian-orbits-2-3.h5"

    with File(gw_file, "r") as gwf:

        data = Data.from_gws(gw_file, orbit_file)
        X1.build(**data.args)(data.measurements)
        assert data.fs == gwf.attrs["fs"]
        for mosa in Data.MOSAS:
            assert data.measurements[f"isi_{mosa}"] != approx(0.0)
            assert data.measurements[f"isi_sb_{mosa}"] != approx(0.0)
            assert data.measurements[f"tmi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_sb_{mosa}"] == approx(0.0)

        data = Data.from_gws(gw_file, orbit_file, gw_dataset="tcb")
        X1.build(**data.args)(data.measurements)
        assert data.fs == gwf.attrs["fs"]
        for mosa in Data.MOSAS:
            assert data.measurements[f"isi_{mosa}"] != approx(0.0)
            assert data.measurements[f"isi_sb_{mosa}"] != approx(0.0)
            assert data.measurements[f"tmi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_sb_{mosa}"] == approx(0.0)

        data = Data.from_gws(gw_file, orbit_file, gw_dataset="tps")
        X1.build(**data.args)(data.measurements)
        assert data.fs == gwf.attrs["fs"]
        for mosa in Data.MOSAS:
            assert data.measurements[f"isi_{mosa}"] != approx(0.0)
            assert data.measurements[f"isi_sb_{mosa}"] != approx(0.0)
            assert data.measurements[f"tmi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_{mosa}"] == approx(0.0)
            assert data.measurements[f"rfi_sb_{mosa}"] == approx(0.0)


def test_lisanode_file_1_4():
    """Test that we can read LISANode files v1.4.

    The test LISANode file can generated using LISANode v1.4, compiling with
    default configuration using

        lisanode run --build lisanode:LISA

    and then running (with LISA Orbits 2.0),

        ./LISA --duration 300.0 --time-origin 2073211130.8175 --orbit-path
        'esa-trailing-orbits.h5' --output 'lisanode-1-4.h5'

    """
    lisanode_file = "tests/lisanode-1-4.h5"

    with File(lisanode_file, "r") as lisanodef:

        data = Data.from_lisanode(lisanode_file)
        X1.build(**data.args)(data.measurements)
        assert data.fs == 1.0 / lisanodef["mpr_12"].attrs["dt"]
        for mosa in Data.MOSAS:
            assert all(
                data.measurements[f"isi_{mosa}"]
                == lisanodef[f"isi_c_fluctuations_{mosa}"]
            )
            assert all(
                data.measurements[f"isi_sb_{mosa}"]
                == lisanodef[f"isi_sb_fluctuations_{mosa}"]
            )
            assert all(
                data.measurements[f"tmi_{mosa}"]
                == lisanodef[f"tmi_c_fluctuations_{mosa}"]
            )
            assert all(
                data.measurements[f"rfi_{mosa}"]
                == lisanodef[f"rfi_c_fluctuations_{mosa}"]
            )
            assert all(
                data.measurements[f"rfi_sb_{mosa}"]
                == lisanodef[f"rfi_sb_fluctuations_{mosa}"]
            )



#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Tests for `TDICombination` class.

Authors:
    Martin Staab <martin.staab@aei.mpg.de>
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
"""

import numpy as np
from pytest import approx, raises

from pytdi import TDICombination

toymodel = {
    "y_12": [(-1, []), (1, ["D_21"])],
    "y_21": [(1, []), (-1, ["D_12"])],
}


def test_init():
    """Check components of a combination are correctly initialized."""
    combination = TDICombination(toymodel)
    assert combination.components == toymodel


def test_measurements():
    """Check that `measurements` returns the correct set of measurements."""
    combination = TDICombination(
        {
            "y_12": [(-1, []), (1, ["D_21", "A_21"])],
            "y_21": [(1, []), (-1, ["D_12", "A_12"])],
        }
    )
    assert combination.measurements == set(["y_12", "y_21"])


def test_delays():
    """Check that `delays` returns the correct set of delay operators."""
    combination = TDICombination(
        {
            "y_12": [(-1, []), (1, ["D_21", "A_21"])],
            "y_21": [(1, []), (-1, ["D_12", "A_12"])],
        }
    )
    assert combination.delays == set(["D_12", "D_21"])


def test_advancements():
    """Check that `advancements` returns the correct set of advancement operators."""
    combination = TDICombination(
        {
            "y_12": [(-1, []), (1, ["D_21", "A_21"])],
            "y_21": [(1, []), (-1, ["D_12", "A_12"])],
        }
    )
    assert combination.advancements == set(["A_12", "A_21"])


def test_transformed():
    """Check that `transformed()` returns a new combination with correct mapping."""
    combination = TDICombination(toymodel)
    with raises(ValueError):
        combination.transformed({})
        combination.transformed({1: 2})
        combination.transformed({1: 1, 2: 1})
    assert combination.transformed({1: 1, 2: 2}) == combination
    assert combination.transformed({1: 2, 2: 1}).components == {
        "y_21": [(-1, []), (1, ["D_12"])],
        "y_12": [(1, []), (-1, ["D_21"])],
    }


def test_simplify():
    """Check that `simplified()` returns an equivalent simplified combination."""
    assert TDICombination(toymodel).simplified().components == {
        "y_12": [(-1, []), (1, ["D_21"])],
        "y_21": [(-1, ["D_12"]), (1, [])],
    }

    combination = TDICombination(
        {
            "y_12": [
                (1.0, []),
                (2.0, []),
                (-3.5, ["A_21", "D_12"]),
                (0.0, ["D_12", "D_21"]),
            ],
            "y_21": [],
        }
    )
    assert combination.simplified().components == {"y_12": [(-0.5, [])]}


def test_add_tdi_combination():
    """Check that addition of two combination is the merger of their components.."""
    combination1 = TDICombination({"y_12": [(1.0, [])]})
    combination2 = TDICombination({"y_12": [(2.0, [])]})
    combination3 = TDICombination({"y_12": [(1.0, ["D_12"])]})
    combination4 = TDICombination({"x_12": [(1.0, ["D_21"])]})
    assert (combination1 + combination2).components == {"y_12": [(3.0, [])]}
    assert (combination1 + combination3).components == {
        "y_12": [(1.0, []), (1.0, ["D_12"])]
    }
    assert (combination1 + combination4).components == {
        "x_12": [(1.0, ["D_21"])],
        "y_12": [(1.0, [])],
    }


def test_subtract():
    """Check subtraction of two combinations."""
    combination1 = TDICombination({"y_12": [(1.0, [])]})
    combination2 = TDICombination({"y_12": [(2.0, [])]})
    combination3 = TDICombination({"y_12": [(1.0, ["D_12"])]})
    combination4 = TDICombination({"x_12": [(1.0, ["D_21"])]})
    assert combination1 - combination1 == TDICombination({})
    assert combination2 - combination2 == TDICombination({})
    assert combination3 - combination3 == TDICombination({})
    assert combination4 - combination4 == TDICombination({})
    assert (combination1 - combination2).components == {"y_12": [(-1.0, [])]}
    assert (combination1 - combination3).components == {
        "y_12": [(-1.0, ["D_12"]), (1.0, [])]
    }
    assert (combination1 - combination4).components == {
        "x_12": [(-1.0, ["D_21"])],
        "y_12": [(1.0, [])],
    }


def test_multiplication():
    """Check multiplication of a combination by a scalar and delays."""
    combination = TDICombination({"y_12": [(1.0, [])]})
    assert 1 * combination == combination
    assert 0 * combination == TDICombination({})
    assert (-1 * combination).components == {"y_12": [(-1.0, [])]}
    assert (["D_12", "D_21"] * combination).components == {
        "y_12": [(1.0, ["D_12", "D_21"])]
    }
    assert ((-1.0, ["D_12", "D_21"]) * combination).components == {
        "y_12": [(-1.0, ["D_12", "D_21"])]
    }


def test_neg():
    """Check opposite of combination."""
    combination = TDICombination(toymodel)
    assert (-combination) + combination == TDICombination({})
    assert (-combination).components == {
        "y_12": [(-1, ["D_21"]), (1, [])],
        "y_21": [(-1, []), (1, ["D_12"])],
    }


def test_composition():
    """Check composition (chaining) of combinations."""
    combination = TDICombination({"y_12": [(2.0, []), (1.0, ["D_21"])]})
    combination_y12 = TDICombination({"x_12": [(3.0, ["D_12"])]})
    assert (combination @ {"y_12": combination_y12}).components == {
        "x_12": [(3.0, ["D_21", "D_12"]), (6.0, ["D_12"])]
    }


def test_build_with_constant_delays():
    """Check computation of a combination with constant delays."""

    fs = 1.0
    size = 1000
    combination = TDICombination({"y_12": [(1, ["A_21"]), (-2, ["D_12"])]})

    def func_measurement(time):
        f_sin = 0.1
        return np.sin(2 * np.pi * f_sin * time)

    times = np.arange(size) / fs
    y12 = func_measurement(times)
    d12 = 8.765

    built = combination.build({"d_12": d12}, fs)
    result = built({"y_12": y12}, unit="phase")
    expected = func_measurement(times + d12) - 2 * func_measurement(times - d12)
    assert result[50:-50] == approx(expected[50:-50])


def test_build_with_variable_delays():
    """Check computation of a combination with variable delays."""

    fs = 1.0
    size = 1000
    combination = TDICombination({"y_12": [(1, []), (-2, ["D_12"])]})

    def func_measurement(time):
        f_sin = 0.1
        return np.sin(2 * np.pi * f_sin * time)

    def func_delays(time):
        slope, offset = 0.1, 0.0
        return slope * time + offset

    times = np.arange(size) / fs
    y12 = func_measurement(times)
    d12 = func_delays(times)

    built = combination.build({"d_12": d12}, fs)
    result = built({"y_12": y12}, unit="phase")
    expected = func_measurement(times) - 2 * func_measurement(
        times - func_delays(times)
    )
    assert result[50:-50] == approx(expected[50:-50])


def test_build_with_variable_delays_and_advancements():
    """Check computation of a combination with variable delays and advancements."""

    fs = 1.0
    size = 1000
    combination = TDICombination({"y_12": [(1, []), (-2, ["A_21", "D_12"])]})

    def func_measurement(time):
        f_sin = 0.1
        return np.sin(2 * np.pi * f_sin * time)

    def func_delays(time):
        slope, offset = 0.1, 0.0
        return slope * time + offset

    times = np.arange(size) / fs
    data = func_measurement(times)
    d12 = func_delays(times)

    built = combination.build({"d_12": d12}, fs, maxiter=100)
    result = built({"y_12": data}, unit="phase")
    expected = -func_measurement(times)
    assert result[50:-150] == approx(expected[50:-150])


def test_build_with_linear_delays_constant_delay_derivatives():
    """Check computation of a combination with linear delays and constant derivatives."""

    fs = 1.0
    size = 1000
    combination = TDICombination({"y_12": [(1, ["A_21"]), (-2, ["D_12"])]})

    def func_measurement(time):
        f_sin = 0.1
        return np.sin(2 * np.pi * f_sin * time)

    offset, slope = 0.0, 0.1

    def func_delays(time):
        return offset + slope * time

    def func_advancements(time):
        return (offset + slope * time) / (1 - slope)

    times = np.arange(size) / fs
    y12 = func_measurement(times)
    d12 = func_delays(times)

    built = combination.build(
        {"d_12": d12}, fs, delay_derivatives={"d_12": slope}, maxiter=100
    )
    result = built({"y_12": y12}, unit="frequency")
    expected = (1 + slope / (1 - slope)) * func_measurement(
        times + func_advancements(times)
    ) - 2 * (1 - slope) * func_measurement(times - func_delays(times))
    assert result[150:-150] == approx(expected[150:-150])


def test_build_with_variable_delays_variable_delay_derivatives():
    """Check computation of a combination with variable delays and variable derivatives."""

    fs = 1.0
    size = 1000
    combination = TDICombination({"y_12": [(1, ["A_21"]), (-2, ["D_12"])]})

    def func_measurement(time):
        fsig = 0.1
        return np.sin(2 * np.pi * fsig * time)

    offset, acceleration = 5.0, 0.00001

    def func_delays(time):
        return offset + acceleration * time**2 / 2

    def func_delay_derivatives(time):
        return acceleration * time

    def func_advancements(time):
        return (
            1
            - acceleration * time
            - np.sqrt(1 - 2 * offset * acceleration - 2 * acceleration * time)
        ) / acceleration

    def func_advancement_derivatives(time):
        return 1 / np.sqrt(1 - 2 * offset * acceleration - 2 * acceleration * time) - 1

    times = np.arange(size) / fs
    y12 = func_measurement(times)
    d12 = func_delays(times)
    dotd12 = func_delay_derivatives(times)

    built = combination.build(
        {"d_12": d12}, fs, delay_derivatives={"d_12": dotd12}, maxiter=100
    )
    result = built({"y_12": y12}, unit="frequency")
    expected = (1 + func_advancement_derivatives(times)) * func_measurement(
        times + func_advancements(times)
    ) - 2 * (1 - func_delay_derivatives(times)) * func_measurement(
        times - func_delays(times)
    )
    assert result[150:-150] == approx(expected[150:-150])



#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Tests for `from_string` method in `TDICombination` class.

Authors:
    Olaf Hartwig <olaf.hartwig@aei.mpg.de>
"""

from pytdi import TDICombination

tdiX1 = {
    "eta_21": [(-1, ["D_12"]), (1, ["D_13", "D_31", "D_12"])],
    "eta_12": [(-1, []), (1, ["D_13", "D_31"])],
    "eta_31": [(-1, ["D_12", "D_21", "D_13"]), (1, ["D_13"])],
    "eta_13": [(-1, ["D_12", "D_21"]), (1, [])],
}

tdiX2 = {
    "eta_21": [
        (-1, ["D_13", "D_31", "D_12"]),
        (-1, ["D_13", "D_31", "D_12", "D_21", "D_12"]),
        (1, ["D_12"]),
        (1, ["D_12", "D_21", "D_13", "D_31", "D_13", "D_31", "D_12"]),
    ],
    "eta_12": [
        (-1, ["D_13", "D_31"]),
        (-1, ["D_13", "D_31", "D_12", "D_21"]),
        (1, []),
        (1, ["D_12", "D_21", "D_13", "D_31", "D_13", "D_31"]),
    ],
    "eta_31": [
        (-1, ["D_13"]),
        (-1, ["D_13", "D_31", "D_12", "D_21", "D_12", "D_21", "D_13"]),
        (1, ["D_12", "D_21", "D_13"]),
        (1, ["D_12", "D_21", "D_13", "D_31", "D_13"]),
    ],
    "eta_13": [
        (-1, []),
        (-1, ["D_13", "D_31", "D_12", "D_21", "D_12", "D_21"]),
        (1, ["D_12", "D_21"]),
        (1, ["D_12", "D_21", "D_13", "D_31"]),
    ],
}


def test_empty():
    """Test if empty string returns empty combination"""
    combination = TDICombination.from_string("")
    assert combination.components == {}


def test_tdi_x1():
    """Check we build TDI X1 correctly."""
    combination = TDICombination.from_string("12131 -12131")
    assert combination.components == tdiX1


def test_tdi_x1_decomposed():
    """Check we build TDI X1 correctly with a decomposed string."""
    combination = TDICombination.from_string("12 21 13 31 -12 -21 -13 -31")
    assert combination.components == tdiX1


def test_tdi_x2():
    """Check we build TDI X2 correctly."""
    combination = TDICombination.from_string("121313121 -131212131")
    assert combination.components == tdiX2


def test_tdi_x2_decomposed():
    """Check we build TDI X2 correctly with a decomposed string."""
    combination = TDICombination.from_string("12 21 13 31 13 31 12 21 -131212131")
    assert combination.components == tdiX2



#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Test that Michelson combinations can reduce laser noise (using LISANode data).
"""

import numpy as np
import scipy
from integration import (
    BASE_PATH,
    COEFFS_AA,
    filter_tf_derivative,
    psd_model,
    psd_threshold,
    read_from_lisanode,
)

import pytdi.intervar
import pytdi.michelson
import pytdi.naming


def test_laser_noise_reduction_with_michelson_1():
    """Test that first-generation Michelson combinations reduce laser noise."""
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-threelasers-lasernoise-staticorbits.h5"
    )
    data_eta = {}
    for i, j in pytdi.naming.lisa_indices():
        name_eta = pytdi.naming.join_indices("eta", i, j)
        name_isc = pytdi.naming.join_indices("isi", i, j)
        data_eta[name_eta] = data_ln[name_isc]
    tdivars = [pytdi.michelson.X1_ETA, pytdi.michelson.Y1_ETA, pytdi.michelson.Z1_ETA]
    names = ["X1_ETA", "Y1_ETA", "Z1_ETA"]
    for tdi, name in zip(tdivars, names):
        data_x = tdi.build(delays, 3.2)(data_eta, unit="phase")
        psd_threshold(
            data_x[200:],
            1e-46,
            freq_range=(1e-4, 5e-1),
            fs=3.2,
            name=f"test_laser_noise_reduction_with_michelson_1:{name}",
        )


def test_reduction_to_3_laser_laser_noise_reduction_with_michelson_1():
    """Test that first-generation Michelson combinations reduce to 3 lasers and laser noise."""
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-staticorbits.h5"
    )
    data_xi = {}
    for i, j in pytdi.naming.lisa_indices():
        name_xi = pytdi.naming.join_indices("xi", i, j)
        name_isc = pytdi.naming.join_indices("isi", i, j)
        name_ref = pytdi.naming.join_indices("rfi", i, j)
        data_xi[name_xi] = data_ln[name_isc]
        data_xi[name_ref] = data_ln[name_ref]

    data_eta = {}
    for key, var in pytdi.intervar.ETA_XI_SET.items():
        data_eta[key] = var.build(delays, 3.2)(data_xi, unit="phase")

    tdivars = [pytdi.michelson.X1_ETA, pytdi.michelson.Y1_ETA, pytdi.michelson.Z1_ETA]
    names = ["X1_ETA", "Y1_ETA", "Z1_ETA"]
    for tdi, name in zip(tdivars, names):
        data_x = tdi.build(delays, 3.2)(data_eta, unit="phase")
        psd_threshold(
            data_x[200:],
            1e-46,
            freq_range=(1e-4, 5e-1),
            fs=3.2,
            name=f"test_reduction_to_3_laser_laser_noise_reduction_with_michelson_1:{name}",
        )


def test_michelson_1():
    """Test that the entire first-generation Michelson combinations reduces laser noise."""
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-staticorbits.h5"
    )
    tdivars = [pytdi.michelson.X1, pytdi.michelson.Y1, pytdi.michelson.Z1]
    names = ["X1", "Y1", "Z1"]
    for tdi, name in zip(tdivars, names):
        data_x = tdi.build(delays, 3.2)(data_ln, unit="phase")
        psd_threshold(
            data_x[200:],
            1e-46,
            freq_range=(1e-4, 5e-1),
            fs=3.2,
            name=f"test_michelson_1:{name}",
        )


def test_factorized_michelson_1():
    """Test that the entire factorized first-generation Michelson combinations reduces laser noise."""
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-staticorbits.h5"
    )
    data = pytdi.interface.Data(data_ln, delays, fs=3.2)
    etas = pytdi.intervar.compute_etas(data, unit="phase")
    names = ["X1", "Y1", "Z1"]
    for i, name in enumerate(names):
        data_x = pytdi.michelson.compute_factorized_michelson(
            data, etas, rot=i, generation=1, unit="phase"
        )
        psd_threshold(
            data_x[200:],
            1e-46,
            freq_range=(1e-4, 5e-1),
            fs=3.2,
            name=f"test_factorized_michelson_1:{name}",
        )


def flexing_filtering_model(freq, arm, delays, fs):
    """Return model for flexing-filtering coupling in the second-generation Michelson combinations"""
    omega = 2 * np.pi * freq
    # Calculate mean delays and doppler velocities
    # times = np.arange(len(next(iter(delays)))) / fs
    times = np.arange(len(delays["d_12"])) / fs
    doppler1, tt1, _, _, _ = scipy.stats.linregress(times, delays[arm[0][0]])
    doppler1_p, tt1_p, _, _, _ = scipy.stats.linregress(times, delays[arm[0][1]])
    doppler2, tt2, _, _, _ = scipy.stats.linregress(times, delays[arm[1][0]])
    doppler2_p, tt2_p, _, _, _ = scipy.stats.linregress(times, delays[arm[1][1]])
    doppler1s2 = doppler1**2 + doppler1_p**2
    doppler2s2 = doppler2**2 + doppler2_p**2
    kf2 = filter_tf_derivative(freq, COEFFS_AA, fs)

    tts1 = tt1 + tt1_p
    tts2 = tt2 + tt2_p
    return (
        2
        * omega**2
        * 1e-26
        * kf2
        * (
            (
                2
                - 2 * np.cos(omega * tts1)
                + np.cos(omega * tts2)
                - 2 * np.cos(omega * (tts1 + tts2))
                + np.cos(omega * (tts2 + 2 * tts1))
            )
            * doppler2s2
            + (
                2
                - 2 * np.cos(omega * tts2)
                + np.cos(omega * tts1)
                - 2 * np.cos(omega * (tts1 + tts2))
                + np.cos(omega * (tts1 + 2 * tts2))
            )
            * doppler1s2
        )
    )


def test_michelson_2():
    """Test that the entire second-generation Michelson combinations reduces laser noise."""
    fs = 3.2
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-keplerianorbits.h5"
    )

    tdivars = [pytdi.michelson.X2, pytdi.michelson.Y2, pytdi.michelson.Z2]
    names = ["X2", "Y2", "Z2"]
    arms = [
        (("d_12", "d_21"), ("d_13", "d_31")),
        (("d_21", "d_12"), ("d_23", "d_32")),
        (("d_31", "d_13"), ("d_32", "d_23")),
    ]
    for tdi, name, arm in zip(tdivars, names, arms):
        data_x = tdi.build(delays, fs)(data_ln, unit="phase")
        model = lambda freq, arm=arm: flexing_filtering_model(freq, arm, delays, fs)
        psd_model(
            data_x[400:],
            model,
            freq_range=(1e-4, 5e-1),
            fs=fs,
            freq_res=1e-4,
            threshold=1e-47,
            name=f"test_michelson_2:{name}",
            confidence=0.99,
        )


def test_factorized_michelson_2():
    """Test that the entire factorized second-generation Michelson combinations reduces laser noise."""
    fs = 3.2
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-keplerianorbits.h5"
    )
    data = pytdi.interface.Data(data_ln, delays, fs=3.2)
    etas = pytdi.intervar.compute_etas(data, unit="phase")
    names = ["X2", "Y2", "Z2"]
    arms = [
        (("d_12", "d_21"), ("d_13", "d_31")),
        (("d_21", "d_12"), ("d_23", "d_32")),
        (("d_31", "d_13"), ("d_32", "d_23")),
    ]
    for i, (name, arm) in enumerate(zip(names, arms)):
        data_x = pytdi.michelson.compute_factorized_michelson(
            data, etas, rot=i, generation=2, unit="phase"
        )
        model = lambda freq, arm=arm: flexing_filtering_model(freq, arm, delays, fs)
        psd_model(
            data_x[400:],
            model,
            freq_range=(1e-4, 5e-1),
            fs=fs,
            freq_res=1e-4,
            threshold=1e-47,
            name=f"test_factorized_michelson_2:{name}",
            confidence=0.99,
        )


def test_michelson_2_doppler():
    """Test that the entire second-generation Michelson combinations reduces laser noise."""
    fs = 3.2
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-keplerianorbits-doppler.h5"
    )
    delay_derivatives = {}
    for delay in delays:
        delay_derivatives[delay] = np.gradient(delays[delay], 1 / fs)

    tdivars = [pytdi.michelson.X2, pytdi.michelson.Y2, pytdi.michelson.Z2]
    names = ["X2", "Y2", "Z2"]
    arms = [
        (("d_12", "d_21"), ("d_13", "d_31")),
        (("d_21", "d_12"), ("d_23", "d_32")),
        (("d_31", "d_13"), ("d_32", "d_23")),
    ]
    for tdi, name, arm in zip(tdivars, names, arms):
        data_x = tdi.build(delays, fs, delay_derivatives=delay_derivatives)(
            data_ln, unit="frequency"
        )
        model = lambda freq, arm=arm: flexing_filtering_model(freq, arm, delays, fs)
        psd_model(
            data_x[400:],
            model,
            freq_range=(1e-4, 5e-1),
            fs=fs,
            freq_res=1e-4,
            threshold=1e-47,
            name=f"test_michelson_2_doppler:{name}",
            confidence=0.99,
        )


def test_factorized_michelson_2_doppler():
    """Test that the entire factorized second-generation Michelson combinations reduces laser noise."""
    fs = 3.2
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-keplerianorbits-doppler.h5"
    )
    data = pytdi.interface.Data(data_ln, delays, fs=3.2)
    etas = pytdi.intervar.compute_etas(data, unit="frequency")
    names = ["X2", "Y2", "Z2"]
    arms = [
        (("d_12", "d_21"), ("d_13", "d_31")),
        (("d_21", "d_12"), ("d_23", "d_32")),
        (("d_31", "d_13"), ("d_32", "d_23")),
    ]
    for i, (name, arm) in enumerate(zip(names, arms)):
        data_x = pytdi.michelson.compute_factorized_michelson(
            data, etas, rot=i, generation=2, unit="frequency"
        )
        model = lambda freq, arm=arm: flexing_filtering_model(freq, arm, delays, fs)
        psd_model(
            data_x[400:],
            model,
            freq_range=(1e-4, 5e-1),
            fs=fs,
            freq_res=1e-4,
            threshold=1e-47,
            name=f"test_factorized_michelson_2_doppler:{name}",
            confidence=0.99,
        )



#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Defines TDI first-and-second-generation Michelson variables.

Authors:
    Martin Staab <martin.staab@aei.mpg.de>
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
    Olaf Hartwig <olaf.hartwig@aei.mpg.de>
"""

from . import core, intervar

#: First-generation :math:`X_1` Michelson combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
X1_ETA = core.LISATDICombination.from_string("12131 -12131")

#: First-generation :math:`Y_1` Michelson combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
Y1_ETA = X1_ETA.rotated()


#: First-generation :math:`Z_1` Michelson combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
Z1_ETA = Y1_ETA.rotated()

#: First-generation :math:`X_1` Michelson combination.
#:
#: This combination is the composition of ``X1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
X1 = X1_ETA @ intervar.ETA_SET

#: First-generation :math:`Y_1` Michelson combination.
#:
#: This combination is the composition of ``Y1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
Y1 = Y1_ETA @ intervar.ETA_SET

#: First-generation :math:`Z_1` Michelson combination.
#:
#: This combination is the composition of ``Z1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
Z1 = Z1_ETA @ intervar.ETA_SET


#: Second-generation :math:`X_2` Michelson combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
X2_ETA = core.LISATDICombination.from_string("131212131 -121313121")

#: Second-generation :math:`Y_2` Michelson combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
Y2_ETA = X2_ETA.rotated()

#: Second-generation :math:`Z_2` Michelson combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
Z2_ETA = Y2_ETA.rotated()

#: Second-generation :math:`X_2` Michelson combination.
#:
#: This combination is the composition of ``X2_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
X2 = X2_ETA @ intervar.ETA_SET

#: Second-generation :math:`Y_2` Michelson combination.
#:
#: This combination is the composition of ``Y2_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
Y2 = Y2_ETA @ intervar.ETA_SET

#: Second-generation :math:`Z_2` Michelson combination.
#:
#: This combination is the composition of ``Z2_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
Z2 = Z2_ETA @ intervar.ETA_SET


def compute_factorized_michelson(
    data, etas=None, *, rot=0, order=45, generation=2, unit="frequency"
):
    r"""Numerically-optimized computation of the Michelson variables.

    These combinations can be factorized and evaluated in multiple steps. This
    reduces computation time (less delays to compute) and in some instances
    also improves numerical precision if intermediary results are already small.

    For first-generation combinations, we can write

    .. math::

        X_1 ={}& (1 - \delay_{121})(\eta_{13} + \delay_{13} \eta_{31}) \\
        &- (1 - \delay_{131}) (\eta_{12} + \delay_{12} \eta_{21}).

    Second-generations combinations can instead be written as

    .. math::

        X_2 ={}& (1 - \delay_{121} - \delay_{12131} + \delay_{1312121})
        (\eta_{13} + \delay_{13} \eta_{31}) \\
        &- (1 - \delay_{131} - \delay_{13121} + \delay_{1213131})
        (\eta_{12} + \delay_{12} \eta_{21}).

    We first define the single round-trips (transponder),

    .. math::

        a_1 = \eta_{12} + \delay_{12} \eta_{21},
        a2 = \eta_{13} + \delay_{13} \eta_{31}.

    In the case of X1, we directly compute the result

    .. math:: (\delay_{131} - 1) a_1 - (\delay_{121} - 1) a_2.

    For X2, we first compute two-arm roundtrip

    .. math::

        r_1 = a_1 + \delay_{121} a_2,
        r_2 = \delay_{131} a_1 + a_2,

    and then compute the final expression

    .. math:: (\delay_{13121} - 1) r_1 - (\delay_{12131} - 1) r_2.

    Note that the signs in these combinations are chosen to agree with the
    existing conventions in PyTDI and the LDC.

    Args:
        data (pytdi.interface.Data): data object
        etas (np.ndarray): pre-computed :math:`\eta_{ij}` intermediary variables
        rot (int): rotation of indices for :math:`X, Y, Z`
        order (int): interpolation order
        unit (str): unit of the input data, either 'frequency' or 'phase'

    Returns:
        (np.ndarray) Evaluated TDI Michelson combination.
    """
    # Define rotations of the variables
    # We can't rely on PyTDIs built-in rotation as
    # the variable names are incompatible
    if rot == 0:
        links = ["12", "21", "13", "31"]
    elif rot == 1:
        links = ["23", "32", "21", "12"]
    elif rot == 2:
        links = ["31", "13", "32", "23"]
    else:
        raise ValueError(f"rot must be 0, 1 or 2 (got {rot})")

    # IFOs measuring single transponder
    # Use pre-computed etas if provided
    if etas is not None:
        x_arm_1 = core.LISATDICombination(
            {f"eta_{links[0]}": [(1, [])], f"eta_{links[1]}": [(1, [f"D_{links[0]}"])]}
        ).build(**data.args)(etas, order=order, unit=unit)
        x_arm_2 = core.LISATDICombination(
            {f"eta_{links[2]}": [(1, [])], f"eta_{links[3]}": [(1, [f"D_{links[2]}"])]}
        ).build(**data.args)(etas, order=order, unit=unit)
    else:
        # Use ETA_SET to compute result from IFO measurements
        x_arm_1 = (
            core.LISATDICombination(
                {
                    f"eta_{links[0]}": [(1, [])],
                    f"eta_{links[1]}": [(1, [f"D_{links[0]}"])],
                }
            )
            @ intervar.ETA_SET
        ).build(**data.args)(data.measurements, order=order, unit=unit)
        x_arm_2 = (
            core.LISATDICombination(
                {
                    f"eta_{links[2]}": [(1, [])],
                    f"eta_{links[3]}": [(1, [f"D_{links[2]}"])],
                }
            )
            @ intervar.ETA_SET
        ).build(**data.args)(data.measurements, order=order, unit=unit)

    if generation == 1:
        # Full combination
        result = core.LISATDICombination(
            {
                "x_arm_1": [(-1, []), (1, [f"D_{links[2]}", f"D_{links[3]}"])],
                "x_arm_2": [(1, []), (-1, [f"D_{links[0]}", f"D_{links[1]}"])],
            }
        ).build(**data.args)(
            {"x_arm_1": x_arm_1, "x_arm_2": x_arm_2},
            order=order,
            unit=unit,
        )
    elif generation == 2:
        # IFOs measuring full roundtrip along both arms
        roundtrip_1 = core.LISATDICombination(
            {"arm_a": [(1, [])], "arm_b": [(1, [f"D_{links[0]}", f"D_{links[1]}"])]}
        ).build(**data.args)(
            {"arm_a": x_arm_1, "arm_b": x_arm_2}, order=order, unit=unit
        )
        roundtrip_2 = core.LISATDICombination(
            {"arm_a": [(1, [f"D_{links[2]}", f"D_{links[3]}"])], "arm_b": [(1, [])]}
        ).build(**data.args)(
            {"arm_a": x_arm_1, "arm_b": x_arm_2}, order=order, unit=unit
        )

        # Full combination
        result = core.LISATDICombination(
            {
                "roundtrip_a": [
                    (-1, []),
                    (
                        1,
                        [
                            f"D_{links[2]}",
                            f"D_{links[3]}",
                            f"D_{links[0]}",
                            f"D_{links[1]}",
                        ],
                    ),
                ],
                "roundtrip_b": [
                    (1, []),
                    (
                        -1,
                        [
                            f"D_{links[0]}",
                            f"D_{links[1]}",
                            f"D_{links[2]}",
                            f"D_{links[3]}",
                        ],
                    ),
                ],
            }
        ).build(**data.args)(
            {"roundtrip_a": roundtrip_1, "roundtrip_b": roundtrip_2},
            order=order,
            unit=unit,
        )
    else:
        raise ValueError(f"invalid generation '{generation}', must be 1 or 2")

    return result



#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Test time-shifting functions (delays and time advancements).
"""

import itertools

import numpy as np
from pytest import approx

from pytdi.dsp import timeshift


def test_constant_integer_timeshift():
    """Test `time_shift()` using constant integer time shifts."""
    data = np.random.normal(size=10)

    shifts = [-2, 2, 0, 10, 11]
    fss = [1.0, 2.0, 11.0]
    orders = [1, 3, 31, 111]

    for shift, fs, order in itertools.product(shifts, fss, orders):
        shifted = timeshift(data, shift * fs, order=order)
        shift_index = int(shift * fs)
        if shift < 0:
            assert np.all(shifted[:-shift_index] == 0)
            assert np.all(shifted[-shift_index:] == data[:shift_index])
        elif shift > 0:
            assert np.all(shifted[-shift_index:] == 0)
            assert np.all(shifted[:-shift_index] == data[shift_index:])
        else:
            assert np.all(shifted == data)


def test_constant_fractional_timeshift_first_order():
    """Test `time_shift()` at first order using a constant time shift."""
    size = 10

    slopes = [1.23]
    offsets = [4.56]
    fss = [1]

    for slope, offset, fs in itertools.product(slopes, offsets, fss):
        times = np.arange(size) / fs
        data = slope * times + offset
        shift = np.random.uniform(low=-size / fs, high=size / fs)
        shifted = timeshift(data, shift * fs, order=1)

        i = np.arange(size)
        k = np.floor(shift * fs).astype(int)
        valid_mask = np.logical_and(i + k > 0, i + k + 1 < size)

        assert np.all(
            shifted[valid_mask] == approx(slope * (times + shift)[valid_mask] + offset)
        )


def test_constant_fractional_timeshift():
    """Test `time_shift()` at higher order using a constant time shift."""
    size = 10

    funcs = [lambda time, fs: np.sin(2 * np.pi * fs / 4 * time)]
    fss = [1, 0.2]
    orders = [11, 31, 101]

    for func, fs, order in itertools.product(funcs, fss, orders):
        times = np.arange(size) / fs
        data = func(times, fs)
        shift = np.random.uniform(low=-size / fs, high=size / fs)
        shifted = timeshift(data, shift * fs, order=order)

        i = np.arange(size)
        k = np.floor(shift * fs).astype(int)
        p = (order + 1) // 2  # pylint: disable=invalid-name
        valid_mask = np.logical_and(i + k - (p - 1) > 0, i + k + p < size)

        assert np.all(
            shifted[valid_mask] == approx(func((times + shift)[valid_mask], fs))
        )


def test_variable_integer_timeshift():
    """Test `time_shift()` using variable integer time shifts."""
    size = 10

    data = np.random.normal(size=size)
    shifts = [
        np.arange(size),
        -2 * np.arange(size) + size // 2,
        -1 * np.ones(size, dtype=int),
    ]
    fss = [1.0, 2.0, 5.0]
    orders = [1, 3, 11, 31]

    for shift, fs, order in itertools.product(shifts, fss, orders):
        shifted = timeshift(data, shift * fs, order=order)
        indices = np.arange(size) + (shift * fs).astype(int)
        zeros_mask = np.logical_or(indices >= size, indices < 0)
        non_zeros_mask = np.invert(zeros_mask)

        assert np.all(shifted[zeros_mask] == 0)
        assert np.all(shifted[non_zeros_mask] == data[indices[non_zeros_mask]])


def test_variable_fractional_timeshift_first_order():
    """Test `time_shift()` at first order using variable time shifts."""
    size = 10

    slopes = [1.23]
    offsets = [4.56]
    fss = [1]

    for slope, offset, fs in itertools.product(slopes, offsets, fss):
        times = np.arange(size) / fs
        data = slope * times + offset
        shifts = np.random.uniform(low=-size / fs, high=size / fs, size=size)
        shifted = timeshift(data, shifts * fs, order=1)

        i = np.arange(size)
        k = np.floor(shifts * fs).astype(int)
        valid_mask = np.logical_and(i + k > 0, i + k + 1 < size)

        assert np.all(
            shifted[valid_mask] == approx(slope * (times + shifts)[valid_mask] + offset)
        )


def test_variable_fractional_timeshift():
    """Test `time_shift()` at higher order using variable time shifts."""
    size = 10

    funcs = [lambda time, fs: np.sin(2 * np.pi * fs / 4 * time)]
    fss = [1, 0.2]
    orders = [11, 31, 101]

    for func, fs, order in itertools.product(funcs, fss, orders):
        times = np.arange(size) / fs
        data = func(times, fs)
        shifts = np.random.uniform(low=-size / fs, high=size / fs, size=size)
        shifted = timeshift(data, shifts * fs, order=order)

        i = np.arange(size)
        k = np.floor(shifts * fs).astype(int)
        p = (order + 1) // 2  # pylint: disable=invalid-name
        valid_mask = np.logical_and(i + k - (p - 1) > 0, i + k + p < size)

        assert np.all(
            shifted[valid_mask] == approx(func((times + shifts)[valid_mask], fs))
        )



#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Test that Sagnac combinations can reduce laser noise (using LISANode data).
"""

from integration import BASE_PATH, psd_threshold, read_from_lisanode

import pytdi.intervar
import pytdi.sagnac


def test_sagnac_1():
    """Test that the entire first-generation Sagnac combinations reduces laser noise."""
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-keplerianorbits.h5"
    )
    names = ["alpha-1", "beta-1", "gamma-1"]
    tdivars = [pytdi.sagnac.ALPHA1, pytdi.sagnac.BETA1, pytdi.sagnac.GAMMA1]
    for tdi, name in zip(tdivars, names):
        data_x = tdi.build(delays, 3.2)(data_ln, unit="phase")
        psd_threshold(data_x[200:], 1e-35, freq_range=(1e-4, 5e-1), fs=3.2, name=name)


def test_factorized_sagnac_1():
    """Test that the factorized first-generation Sagnac combinations reduces laser noise."""
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-keplerianorbits.h5"
    )
    data = pytdi.interface.Data(data_ln, delays, fs=3.2)
    etas = pytdi.intervar.compute_etas(data, unit="phase")
    names = ["facto-alpha-1", "facto-beta-1", "facto-gamma-1"]
    for i, name in enumerate(names):
        data_x = pytdi.sagnac.compute_factorized_sagnac(
            data, etas, rot=i, generation=1, unit="phase"
        )
        psd_threshold(data_x[200:], 1e-35, freq_range=(1e-4, 5e-1), fs=3.2, name=name)


def test_symmetric_sagnac_1():
    """Test that the entire first-generation Sagnac combinations reduces laser noise."""
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-keplerianorbits.h5"
    )
    names = ["zeta-1"]
    tdivars = [pytdi.sagnac.ZETA1]
    for tdi, name in zip(tdivars, names):
        data_x = tdi.build(delays, 3.2)(data_ln, unit="phase")
        psd_threshold(data_x[200:], 1e-30, freq_range=(1e-4, 5e-1), fs=3.2, name=name)


def test_sagnac_2():
    """Test that the entire second-generation Sagnac combinations reduces laser noise."""
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-keplerianorbits.h5"
    )
    names = ["alpha-2", "beta-2", "gamma-2"]
    tdivars = [pytdi.sagnac.ALPHA2, pytdi.sagnac.BETA2, pytdi.sagnac.GAMMA2]
    for tdi, name in zip(tdivars, names):
        data_x = tdi.build(delays, 3.2)(data_ln, unit="phase")
        psd_threshold(data_x[200:], 1e-41, freq_range=(1e-4, 5e-1), fs=3.2, name=name)


def test_factorized_sagnac_2():
    """Test that the factorized second-generation Sagnac combinations reduces laser noise."""
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-keplerianorbits.h5"
    )
    data = pytdi.interface.Data(data_ln, delays, fs=3.2)
    etas = pytdi.intervar.compute_etas(data, unit="phase")
    names = ["facto-alpha-2", "facto-beta-2", "facto-gamma-2"]
    for i, name in enumerate(names):
        data_x = pytdi.sagnac.compute_factorized_sagnac(
            data, etas, rot=i, generation=2, unit="phase"
        )
        psd_threshold(data_x[200:], 1e-41, freq_range=(1e-4, 5e-1), fs=3.2, name=name)


def test_symmetric_sagnac_2():
    """Test that the entire second-generation Sagnac combinations reduces laser noise."""
    data_ln, delays = read_from_lisanode(
        BASE_PATH + "lisa-sixlasers-lasernoise-keplerianorbits.h5"
    )
    names = ["zeta-21", "zeta-22", "zeta-23"]
    tdivars = [pytdi.sagnac.ZETA21, pytdi.sagnac.ZETA22, pytdi.sagnac.ZETA23]
    for tdi, name in zip(tdivars, names):
        data_x = tdi.build(delays, 3.2)(data_ln, unit="phase")
        psd_threshold(data_x[200:], 1e-38, freq_range=(1e-4, 5e-1), fs=3.2, name=name)



#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Tests for `LISATDICombination` class.

Authors:
    Martin Staab <martin.staab@aei.mpg.de>
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
"""

from pytdi import LISATDICombination

toymodel = {
    "y_12": [(-1, []), (1, ["D_12"])],
    "y_21": [(1, []), (-1, ["D_21"])],
}


def test_init():
    """Check components of a combination are correctly initialized."""
    combination = LISATDICombination(toymodel)
    assert combination.components == toymodel


def test_addition():
    """Check that components are correctly added together."""
    tdi_comb1 = LISATDICombination({"y_12": [(1, [])]})
    tdi_comb2 = LISATDICombination({"y_12": [(2, [])]})
    tdi_comb3 = LISATDICombination({"y_12": [(1, ["D_12"])]})
    tdi_comb4 = LISATDICombination({"x_23": [(1, ["D_23"])]})
    assert (tdi_comb1 + tdi_comb2).components == {"y_12": [(3, [])]}
    assert (tdi_comb1 + tdi_comb3).components == {"y_12": [(1, []), (1, ["D_12"])]}
    assert (tdi_comb1 + tdi_comb4).components == {
        "x_23": [(1, ["D_23"])],
        "y_12": [(1, [])],
    }


def test_rotatation():
    """Test rotation transformation."""
    combination = LISATDICombination(toymodel)
    assert combination.rotated(0) == combination
    assert combination.rotated() == combination.rotated(1)
    assert combination.rotated(1).components == {
        "y_23": [(-1, []), (1, ["D_23"])],
        "y_32": [(1, []), (-1, ["D_32"])],
    }
    assert combination.rotated(2).components == {
        "y_31": [(-1, []), (1, ["D_31"])],
        "y_13": [(1, []), (-1, ["D_13"])],
    }
    assert combination.rotated(3) == combination


def test_reflection():
    """Test reflection transformation."""
    combination = LISATDICombination(
        {
            "y_12": [(-1, []), (5, ["D_12"])],
            "y_23": [(-0.5, []), (1, ["D_32"])],
            "y_31": [(1, []), (-1, ["D_21"])],
        }
    )
    assert combination.reflected(1).components == {
        "y_13": [(-1, []), (5, ["D_13"])],
        "y_32": [(-0.5, []), (1, ["D_23"])],
        "y_21": [(1, []), (-1, ["D_31"])],
    }
    assert combination.reflected(2).components == {
        "y_32": [(-1, []), (5, ["D_32"])],
        "y_21": [(-0.5, []), (1, ["D_12"])],
        "y_13": [(1, []), (-1, ["D_23"])],
    }
    assert combination.reflected(3).components == {
        "y_21": [(-1, []), (5, ["D_21"])],
        "y_13": [(-0.5, []), (1, ["D_31"])],
        "y_32": [(1, []), (-1, ["D_12"])],
    }



#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Test that Michelson combinations can reduce laser noise (using LISANode data).
"""

import h5py
import matplotlib.pyplot as plt
import numpy as np
import scipy.ndimage
import scipy.signal
import scipy.stats

BASE_PATH = "tests/data/"
COEFFS_AA = np.load(BASE_PATH + "filter_coeffs.npy")


def filter_tf_derivative(freq, coeffs, fs):
    """Calculate squared modulus of derivative of filter transfer function.

    Args:
        freq: frequency bins to evaluate
        coeffs: filter coefficients
        fs: sampling frequency of the data [Hz]

    Returns:
        squared modulus of derivative of filter transer function
    """
    res = 0
    for i, coeff in enumerate(coeffs):
        factor = i / (10 * fs)
        res += 2 * coeff * factor * np.sin(2 * np.pi * freq * factor)
    return res**2


def psd_threshold(data, threshold, freq_range, fs=1, name=None, plotting=True):
    """Check that data is below certain threshold.

    This methods checks that the power spectral density, computed with Welch method,
    remains below a certain threshold.

    We only check the psd inside the given frequency range.

    Args:
        data: array of data given as a time series
        fs: sampling frequency of the data [Hz]
        threshold: constant threshold
        freq_range: couple (min_freq, max_freq) defining the frequency range to check
    """
    # Check that frequencies are consistent
    min_freq, max_freq = freq_range
    assert min_freq >= 0
    assert max_freq <= fs / 2
    # Compute the nperseg so that we reach the min frequency (margin of 10 samples)
    # Check that we have enough data for at least two averages
    nperseg = fs / min_freq + 10
    nseg = len(data) // nperseg
    assert nseg >= 2
    # Compute Welch method, and model
    freq, estimate = scipy.signal.welch(
        data, fs=fs, nperseg=nperseg, noverlap=0, detrend=False, window="nuttall"
    )
    # Only consider frequency range, dc and Nyquist frequency are excluded
    mask = np.logical_and(freq > min_freq, freq < max_freq)
    # Plot results
    if plotting:
        plt.figure(figsize=(16, 8))
        plt.axvline(x=min_freq, ls="--", color="grey")
        plt.axvline(x=max_freq, ls="--", color="grey")
        plt.loglog(freq, estimate, label="psd")
        plt.axhline(y=threshold, color="black", label="threshold")
        plt.xlabel("Frequency (Hz)")
        plt.ylabel("Power spectral density (1/Hz)")
        plt.legend(loc="upper left")
        plt.savefig(name + ".pdf")
        plt.close()

    assert np.all(estimate[mask] < threshold)


def psd_model(
    data,
    model,
    freq_range,
    fs=1,
    freq_res=None,
    threshold=0,
    confidence=0.9,
    name=None,
    plotting=True,
):
    """Check that data follows a power spectral density model.

    This methods checks that the power spectral density of the model, computed with Welch method,
    remains within the confidence interval around the model. If you specify a threshold,
    data below this level is accepted.

    We only check that the data follows the model inside the given frequency range.

    Args:
        data: array of data given as a time series
        fs: sampling frequency of the data [Hz]
        model: function returning the model PSD as a function of frequency in Hz
        freq_range: couple (min_freq, max_freq) defining the frequency range to check
        threshold: noise floor level
        confidence: confidence interval around the model

    Returns:
        True if the data follows the model, otherwise returns False.
    """

    # pylint: disable=too-many-locals
    def model_spectral_leakage(model, window, nperseg, fs):
        win = window(nperseg)
        freq = np.fft.fftshift(np.fft.fftfreq(nperseg, d=1 / fs))
        win_fft = np.fft.fft(win)
        expected = (
            scipy.ndimage.convolve1d(model(freq), np.abs(win_fft) ** 2, mode="wrap")
            / np.sum(win**2)
            / nperseg
        )
        return expected[: nperseg // 2 + 1]

    def calculate_confidence_intervall(estimate, expected, nseg, confidence):
        # Calculate the degrees of freedom
        df = 2 * nseg  # pylint: disable=invalid-name
        # Each element of df*estimate/expected is chi^2-distributed with degrees of freedom df
        # Probability that a single point is outside of the confidence interval
        alpha = 1 - confidence ** (1 / len(estimate))
        # Compute confidence interval
        ratio_lower = scipy.stats.chi2.ppf(alpha / 2, df)
        ratio_upper = scipy.stats.chi2.ppf(1 - alpha / 2, df)
        return expected / df * ratio_lower, expected / df * ratio_upper

    def plot(freq, estimate, expected, intervall, mask_valid, freq_range, threshold):
        freq_min, freq_max = freq_range
        expected_lower, expected_upper = intervall
        mask_out = np.logical_or(
            expected_lower > estimate[mask_valid], expected_upper < estimate[mask_valid]
        )
        plt.figure(figsize=(16, 8))
        plt.axvline(x=freq_min, ls="--", color="grey")
        plt.axvline(x=freq_max, ls="--", color="grey")
        plt.axhline(y=threshold, ls="--", color="grey")
        plt.loglog(freq[mask_valid], expected_lower, ls="--", color="black")
        plt.loglog(freq[mask_valid], expected_upper, ls="--", color="black")
        plt.loglog(freq, estimate, label="psd")
        plt.loglog(freq, expected, color="black", label="model")
        plt.scatter(
            freq[mask_valid][mask_out], estimate[mask_valid][mask_out], color="red"
        )
        plt.xlabel("Frequency (Hz)")
        plt.ylabel("Power spectral density (1/Hz)")
        plt.legend(loc="upper left")
        plt.savefig(name + ".pdf")
        plt.close()

    # Check that frequencies are consistent
    freq_min, freq_max = freq_range
    assert freq_min >= 0
    assert freq_max <= fs / 2
    # Compute the nperseg so that we reach the min frequency (margin of 10 samples)
    # Check that we have enough data for at least two averages
    if freq_res is None:
        nperseg = int(fs / freq_min + 10)
    else:
        nperseg = int(fs / freq_res)
    nseg = len(data) // nperseg
    assert nseg >= 2
    # Compute Welch method, and model
    freq, estimate = scipy.signal.welch(
        data, fs=fs, nperseg=nperseg, noverlap=0, detrend=False, window="nuttall"
    )
    expected = model_spectral_leakage(model, scipy.signal.windows.nuttall, nperseg, fs)
    # Only consider frequency range, dc and Nyquist frequency are excluded
    mask_freq = np.logical_and(freq > freq_min, freq < freq_max)
    mask_floor = estimate > threshold
    mask_valid = np.logical_and(mask_freq, mask_floor)
    intervall = calculate_confidence_intervall(
        estimate[mask_valid], expected[mask_valid], nseg, confidence
    )

    # Plot results
    if plotting:
        plot(freq, estimate, expected, intervall, mask_valid, freq_range, threshold)

    expected_lower, expected_upper = intervall
    assert np.all(expected_lower < estimate[mask_valid])
    assert np.all(expected_upper > estimate[mask_valid])


def read_from_lisanode(path, skipped=0):
    """Read data from LISANode simulations.

    Args:
        path: path to HDF5 simulation product file
        skipped: number of samples to skip
    """
    pm_names = {
        "isi_12": "s_c_fluctuations1",
        "isi_13": "s_c_fluctuations1_p",
        "isi_23": "s_c_fluctuations2",
        "isi_21": "s_c_fluctuations2_p",
        "isi_31": "s_c_fluctuations3",
        "isi_32": "s_c_fluctuations3_p",
        "rfi_12": "ref_c_fluctuations1",
        "rfi_13": "ref_c_fluctuations1_p",
        "rfi_23": "ref_c_fluctuations2",
        "rfi_21": "ref_c_fluctuations2_p",
        "rfi_31": "ref_c_fluctuations3",
        "rfi_32": "ref_c_fluctuations3_p",
        "tmi_12": "tm_c_fluctuations1",
        "tmi_13": "tm_c_fluctuations1_p",
        "tmi_23": "tm_c_fluctuations2",
        "tmi_21": "tm_c_fluctuations2_p",
        "tmi_31": "tm_c_fluctuations3",
        "tmi_32": "tm_c_fluctuations3_p",
    }

    ranging_names = {
        "d_12": "ranging1",
        "d_13": "ranging1_p",
        "d_23": "ranging2",
        "d_21": "ranging2_p",
        "d_31": "ranging3",
        "d_32": "ranging3_p",
    }

    beatnotes = {
        "isi_12": -10,
        "isi_13": 10,
        "isi_23": -10,
        "isi_21": 10,
        "isi_31": -10,
        "isi_32": 10,
        "rfi_12": -10,
        "rfi_13": 10,
        "rfi_23": -10,
        "rfi_21": 10,
        "rfi_31": -10,
        "rfi_32": 10,
        "tmi_12": -10,
        "tmi_13": 10,
        "tmi_23": -10,
        "tmi_21": 10,
        "tmi_31": -10,
        "tmi_32": 10,
    }

    data = {}
    delays = {}
    with h5py.File(path, "r") as hdf5:
        phasemeasurements = {}
        for key, name in pm_names.items():
            phasemeasurements[key] = hdf5[name][skipped:]

        for key, beatnote in beatnotes.items():
            data[key] = phasemeasurements[key] * np.sign(beatnote)

        for key, name in ranging_names.items():
            delays[key] = hdf5[name][skipped:]

    return data, delays



#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Test time-shifting functions (delays and time advancements).
"""

import numpy as np
from pytest import approx

import pytdi.dsp


def test_constant_delay():
    """Test `calculate_advancement()` with constant delays"""
    delay = 10.234
    size = 100
    fs = 1
    order = 5
    delta = 1e-9
    maxiter = 10

    delays = np.full(size, delay)
    advancements = pytdi.dsp.calculate_advancements(
        delays, fs, order=order, delta=delta, maxiter=maxiter
    )

    num_tabs = order + 1
    k = int(delay * fs)
    assert np.all(advancements[: -(k + num_tabs // 2)] == approx(delay))


def test_variable_linear_delay():
    """Test `calculate_advancement()` with variable delays"""
    offset = 10.234
    slope = 0.0321
    size = 100
    fs = 1
    order = 5
    delta = 1e-9
    maxiter = 10

    def delay(times):
        return offset + slope * times

    def advancement(times):
        return (offset + slope * times) / (1 - slope)

    times = np.arange(size) / fs
    delays = delay(times)
    advancements = pytdi.dsp.calculate_advancements(
        delays, fs, order=order, delta=delta, maxiter=maxiter
    )
    advancements_model = advancement(times)

    num_tabs = order + 1
    k_max = np.max(np.floor(advancements_model * fs)).astype(int)
    cut_end = -(k_max + num_tabs // 2)
    assert np.all(advancements[:-cut_end] == approx(advancements_model[:-cut_end]))



from docutils import nodes

from pytdi import TDICombination


def standard_combinations(app, what, name, obj, options, lines):

    # Tweak documentation for pre-defined TDI combinations
    if what == "data" and isinstance(obj, TDICombination):

        lines.append("")
        delays = ", ".join(f"``{d}``" for d in sorted(obj.delays))
        lines.append(f":Delays: {delays}.")

        lines.append("")
        measurements = ", ".join(f"``{m}``" for m in sorted(obj.measurements))
        lines.append(f":Measurements: {measurements}.")

        lines.append(":Components:")
        lines.append("  .. code-block:: python")
        lines.append("")
        lines.append(f"    {obj.components}")


def setup(app):
    app.connect("autodoc-process-docstring", standard_combinations)
